package ru.dbotthepony.kstarbound.defs

import com.google.common.collect.ImmutableList
import com.google.gson.GsonBuilder
import com.google.gson.JsonArray
import com.google.gson.TypeAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.io.CustomEnumTypeAdapter
import ru.dbotthepony.kstarbound.io.IStringSerializable
import ru.dbotthepony.kstarbound.io.Vector2dTypeAdapter
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

enum class JsonFunctionInterpolation(vararg aliases: String) : IStringSerializable {
	LINEAR;

	private val aliases: Array<String> = Array(aliases.size) { aliases[it].uppercase() }

	override fun match(name: String): Boolean {
		for (alias in aliases)
			if (name == alias)
				return true

		return name == this.name
	}

	override fun write(out: JsonWriter) {
		out.value(this.name)
	}

	companion object {
		val ADAPTER: TypeAdapter<JsonFunctionInterpolation> = CustomEnumTypeAdapter(values()).nullSafe()
	}
}

enum class JsonFunctionConstraint(vararg aliases: String) : IStringSerializable {
	CLAMP;

	private val aliases: Array<String> = Array(aliases.size) { aliases[it].uppercase() }

	override fun match(name: String): Boolean {
		for (alias in aliases)
			if (name == alias)
				return true

		return name == this.name
	}

	override fun write(out: JsonWriter) {
		out.value(this.name)
	}

	companion object {
		val ADAPTER: TypeAdapter<JsonFunctionConstraint> = CustomEnumTypeAdapter(values()).nullSafe()
	}
}

class JsonFunction(
	val interpolation: JsonFunctionInterpolation,
	val constraint: JsonFunctionConstraint,
	val ranges: List<Vector2d>
) {
	companion object : TypeAdapter<JsonFunction>() {
		override fun write(out: JsonWriter, value: JsonFunction) {
			TODO("Not yet implemented")
		}

		override fun read(reader: JsonReader): JsonFunction {
			reader.beginArray()

			val interpolation = JsonFunctionInterpolation.ADAPTER.read(reader)
			val constraint = JsonFunctionConstraint.ADAPTER.read(reader)

			val ranges = ArrayList<Vector2d>()

			while (reader.peek() != JsonToken.END_ARRAY) {
				ranges.add(Vector2dTypeAdapter.read(reader))
			}

			reader.endArray()

			return JsonFunction(interpolation, constraint, ImmutableList.copyOf(ranges))
		}

		fun registerGson(gsonBuilder: GsonBuilder) {
			gsonBuilder.registerTypeAdapter(JsonFunctionConstraint::class.java, JsonFunctionConstraint.ADAPTER)
			gsonBuilder.registerTypeAdapter(JsonFunctionInterpolation::class.java, JsonFunctionInterpolation.ADAPTER)
			gsonBuilder.registerTypeAdapter(JsonFunction::class.java, this)
		}
	}
}
